#pragma once

#include <core/global_definitions.hpp> 
#include <core/base_column.hpp>


namespace CoGaDB{
	namespace gpu{		
	

		class GPU_Positionlist{
			public:
			virtual void clear() = 0;
			virtual const shared_pointer_namespace::shared_ptr<GPU_Positionlist> aggregate(shared_pointer_namespace::shared_ptr<GPU_Positionlist> dev_tids) const = 0;
			virtual bool insert(std::vector<TID>* pos_list) = 0;
			virtual size_t size() const throw()=0;
			virtual const std::type_info& type() const throw() = 0;
			virtual const boost::any get(TID tid)=0; 
			virtual const PositionListPtr toHostPositionlist() const = 0;
			virtual void print() const throw() = 0;
			virtual ~GPU_Positionlist();
		};

//		class GPU_2DPositionlist{
//			public:
//			virtual size_t size() const throw()=0;
//			virtual const std::type_info& type() const throw() = 0;
//			virtual const boost::any get(TID tid)=0; 
//			virtual const std::pair<std::vector<TID>*,std::vector<TID>*> toHost_2DPositionlist() const = 0;
//		};

		typedef shared_pointer_namespace::shared_ptr<GPU_Positionlist> GPU_PositionlistPtr;

		class GPU_Base_Column{
			public:
			//typedef shared_pointer_namespace::shared_ptr<GPU_Base_Column> GPU_Base_ColumnPtr;

			GPU_Base_Column(const std::string& name, AttributeType db_type);

			virtual ~GPU_Base_Column();
		
			virtual bool insert(const boost::any& new_Value)=0;
			virtual bool update(TID tid, const boost::any& new_Value)=0;
			virtual bool remove(TID tid)=0;

			virtual const boost::any get(TID tid)=0; //not const, because operator [] does not provide const return type and the child classes rely on []

			/***************** column algebra operations *****************/	
			/*! \brief adds constant to column
			 *  \details for all indeces i holds the following property: B[i]=A[i]+new_Value*/		
			virtual bool add(const boost::any& new_Value) = 0;
			/*! \brief vector addition of two columns
			 *  \details for all indeces i holds the following property: C[i]=A[i]+B[i]*/				
			virtual bool add(shared_pointer_namespace::shared_ptr<GPU_Base_Column> column) = 0;
			/*! \brief substracts constant from column
			 *  \details for all indeces i holds the following property: B[i]=A[i]-new_Value*/		
			virtual bool minus(const boost::any& new_Value) = 0;
			/*! \brief vector substraction of two columns
			 *  \details for all indeces i holds the following property: C[i]=A[i]-B[i]*/		
			virtual bool minus(shared_pointer_namespace::shared_ptr<GPU_Base_Column> column) = 0;	
			/*! \brief multiply constant with column
			 *  \details for all indeces i holds the following property: B[i]=A[i]*new_Value*/	
			virtual bool multiply(const boost::any& new_Value) = 0;
			/*! \brief multiply two columns A and B
			 *  \details for all indeces i holds the following property: C[i]=A[i]*B[i]*/			
			virtual bool multiply(shared_pointer_namespace::shared_ptr<GPU_Base_Column> column) = 0;
			/*! \brief devide values in column by a constant
			 *  \details for all indeces i holds the following property: B[i]=A[i]/new_Value*/				
			virtual bool division(const boost::any& new_Value) = 0;	
			/*! \brief devide column A with column B
			 *  \details for all indeces i holds the following property: C[i]=A[i]/B[i]*/				
			virtual bool division(shared_pointer_namespace::shared_ptr<GPU_Base_Column> column) = 0;	

			virtual void print() const throw()=0;
			virtual size_t size() const throw()=0;
			virtual const std::type_info& type() const throw() = 0;
			
			virtual const shared_pointer_namespace::shared_ptr<GPU_Base_Column> copy() const = 0;
			virtual const shared_pointer_namespace::shared_ptr<GPU_Base_Column> materialize(shared_pointer_namespace::shared_ptr<GPU_Positionlist> pos_list) const = 0;			
			
			/*! \brief returns database type of column (as defined in "SQL" statement)*/
			AttributeType getType() const throw();
			const std::string getName() const throw();

			protected:
			std::string name_;
			AttributeType db_type_;
		};		

		//typedef GPU_Base_Column::GPU_Base_ColumnPtr GPU_Base_ColumnPtr;

		typedef shared_pointer_namespace::shared_ptr<GPU_Base_Column> GPU_Base_ColumnPtr;
		
//		typedef shared_pointer_namespace::shared_ptr<GPU_Positionlist> GPU_PositionlistPtr;

//		//factory function
//		const GPU_Base_Column* createGPUColumn(AttributeType type, const std::string& name);
//		//copy columns from host to device
//		bool copy_column_host_to_device(const std::vector<int>& host_column, GPU_Base_Column* device_column);
//		bool copy_column_host_to_device(const std::vector<float>& host_column, GPU_Base_Column* device_column);	
//		//
//		//GPU_PositionList* copy_positionlist_host_to_device(std::vector<TID>*);
//			
//		//copy columns from device to host	
//		bool copy_column_device_to_host(GPU_Base_Column* device_column, std::vector<int>& host_column); //host columns are not const, because data is inserted in real columns directly 
//		bool copy_column_device_to_host(GPU_Base_Column* device_column, std::vector<float>& host_column);		

		const GPU_Base_ColumnPtr createGPUColumn(AttributeType type, const std::string& name);

		const ColumnPtr copy_column_device_to_host(GPU_Base_ColumnPtr device_column);
		const GPU_Base_ColumnPtr copy_column_host_to_device(ColumnPtr host_column);

		const PositionListPtr copy_PositionList_device_to_host(GPU_PositionlistPtr device_pos_list);
		const GPU_PositionlistPtr copy_PositionList_host_to_device(PositionListPtr pos_list);

		const GPU_Base_ColumnPtr materializeColumn(GPU_Base_ColumnPtr device_column,GPU_PositionlistPtr device_pos_list);

	}; //end namespace gpu
}; //end namespace CogaDB
		
