
#include <iostream>
#include <fstream>
#include <utility>

#include <core/base_table.hpp>
#include <core/table.hpp>
#include <core/gpu_column_cache.hpp>

#include <lookup_table/lookup_table.hpp>

#include <util/filesystem.hpp>
#include <util/time_measurement.hpp>
#include <util/iostream.hpp>
#include <util/reduce_by_keys.hpp>

#include <boost/filesystem.hpp>
#include <boost/any.hpp>
#include <boost/tokenizer.hpp>
#include <boost/lexical_cast.hpp>

#include <boost/iterator/zip_iterator.hpp>
#include <boost/tuple/tuple.hpp>
#include <boost/tuple/tuple_comparison.hpp>

#include <gpu/gpu_base_column.hpp>
#include <gpu/gpu_algorithms.hpp>

#include <query_processing/query_processor.hpp>

using namespace std;

namespace CoGaDB{

	BaseTable::BaseTable(const std::string& name, const TableSchema& schema) : name_(name), schema_(schema){

	}

	BaseTable::~BaseTable(){

	}
	
	const std::string& BaseTable::getName() const throw(){
		return name_;
	}

	const TableSchema BaseTable::getSchema() const throw(){
		return schema_;
	}


	//using namespace std;

//	const TableSchema mergeTableSchemas(const TableSchema& schema1, const std::string& join_attributname_table1,
//													const TableSchema& schema2, const std::string& join_attributname_table2){
	const TableSchema mergeTableSchemas(const TableSchema& schema1, const std::string& ,
													const TableSchema& schema2, const std::string& ){
 		 TableSchema result_schema, schema_(schema1), schema(schema2);
		 TableSchema::iterator it,it2;
		 map<string, pair<AttributeType,string> > schema_map;
		 map<string, pair<AttributeType,string> >::iterator map_it;

		

		 for(it=schema_.begin();it!=schema_.end();it++){
			 schema_map[it->second]=*it;
			 if(!quiet && verbose) cout << "it->second: " << it->second << endl;
			 result_schema.push_back(*it);
		 }

		//TableSchema schema=table->getSchema();
		 for(it2=schema.begin();it2!=schema.end();it2++){
			map_it=schema_map.find(it2->second);
			if(map_it!=schema_map.end()){
				schema_map[it2->second]=*it2;
			   if(!quiet && verbose) cout << "FOUND Duplicate!" << it2->second << endl;
				result_schema.push_back(*it2);	
			}else{
				schema_map[it2->second]=*it2;
		   	if(!quiet && verbose) cout << "it2->second: " << it2->second << endl;
				result_schema.push_back(*it2);				
			}
			
		 }
		 if(!quiet && verbose) cout << "[DEBUG]: JOIN(): schema of new table:" << endl << "| ";
		 if(!quiet && verbose) 
		 for(it=result_schema.begin();it!=result_schema.end();it++)
			cout << (it->second) << " |"; // << endl;

		cout << endl;
		return result_schema;
	}


        const TablePtr BaseTable::createResultTable(TablePtr table, PositionListPtr tids, MaterializationStatus mat_stat, const std::string& operation_name){
                if(!table || !tids) return  TablePtr();
            
                TablePtr result_table;
		if(mat_stat==MATERIALIZE){

			result_table = TablePtr(new Table(operation_name+string("( ")+table->getName()+" )",table->getSchema())) ;//tmp_schema));
			//fetch result tuples
			for(unsigned int i=0;i<tids->size();i++){
				TID j = (*tids)[i];
				result_table->insert( table->fetchTuple(j) );
			}

		}else if(mat_stat==LOOKUP){

			if(!quiet) cout << "create Lookup Table for Table " << table->getName() << " ..." << endl;

			result_table = createLookupTableforUnaryOperation(string("Lookup_Table_")+operation_name+string("(")+table->getName() +" )", table, tids);

		}else{
			cout << "Error! Unkown Materialization Option!" << endl;
			return TablePtr(); ////error, return NULL pointer
		}
		
		return result_table;           
            
        
        }
        

	/***************** status report *****************/
	unsigned int BaseTable::getNumberofRows() const throw(){
		const std::vector<ColumnPtr>& columns_ = this->getColumns();
		if(columns_.size()>0){
			return columns_[0]->size();
		}else{
			return 0;
		}
	}

	/***************** relational operations *****************/
	const TablePtr BaseTable::selection(TablePtr table, const std::string& column_name, const boost::any& value_for_comparison, const ValueComparator& comp, MaterializationStatus mat_stat, ParallelizationMode comp_mode, const ComputeDevice comp_dev){
		if(!table){
			cout << "Fatal Error: in BaseTable::selection(): input table pointer is NULL!" << endl;
			return TablePtr();
		}

			Timestamp begin,end; 
			begin=getTimestamp();
			ColumnPtr column = table->getColumnbyName(column_name);
			if(column.get()==NULL){
				cout << "Error! Could not look up Column: " << column_name << endl;
				exit(-1);
			}
			
		//perform selection
		PositionListPtr tids;
		if(comp_mode==SERIAL){
			if(comp_dev==CPU){
			tids=column->selection(value_for_comparison, comp);
			}else if(comp_dev==GPU){
				cout << "FATAL ERROR! in BaseTable::selection(): incompatible Parameters: ParallelizationMode (SERIAL), ComputeDevice (GPU)" << endl;
				cout << "Exiting..." << endl;
				exit(-1);
			}

		}else if(comp_mode==PARALLEL){
			if(comp_dev==CPU){
				tids=column->parallel_selection(value_for_comparison, comp, boost::thread::hardware_concurrency());
			}else if(comp_dev==GPU){
                               if(!column->isMaterialized()) column=column->materialize();
				//copy original column to GPU RAM
				gpu::GPU_Base_ColumnPtr device_column = GPU_Column_Cache::instance().getGPUColumn(column); //copy_column_host_to_device(*this);
				if(!device_column){
				   cout << "Error! Host to Device copy operation failed!" << endl;
				   cout << "For Column '" << column->getName() << "' of size '" << column->getSizeinBytes() << "Byte' (" << column->size() << " elements of type " << column->type().name() << ")" << endl;
				   cout << "File: " << __FILE__ << " at Line: " << __LINE__ << endl;
				   printGPUStatus();
				   return TablePtr();
				}
				//compute result tid list on GPU
				gpu::GPU_PositionlistPtr device_tids = gpu::GPU_Operators::selection(device_column, value_for_comparison, comp); 
				if(!device_tids){
				   cout << "Error! GPU Operator 'SELECTION' failed!" << endl;
				   cout << "File: " << __FILE__ << " at Line: " << __LINE__ << endl;
				   printGPUStatus();
				   return TablePtr();
				}
				//copy result tid list back to CPU RAM
				tids = copy_PositionList_device_to_host(device_tids);
				if(!tids){
				   cout << "Error! Device to Host copy operation failed!" << endl;
				   cout << "File: " << __FILE__ << " at Line: " << __LINE__ << endl;
				   return TablePtr();
				}
			}				
			
		}else{
			cout << "Error! Invalid Parallelization Mode in BaseTable::selection() " << endl;
			exit(-1);		
		}

		assert(tids!=NULL);
               // cout << "SELECTION: result TID List length: " << tids->size() << "("  << __FILE__ << ")" <<  endl;
                TablePtr result_table = createResultTable(table, tids, mat_stat, "selection");
		return result_table;
	}
        
        const TablePtr BaseTable::selection(TablePtr table, const Disjunction& disjunction, MaterializationStatus mat_stat, ParallelizationMode comp_mode){
            if(!table){
                    cout << "Fatal Error: in BaseTable::selection(): input table pointer is NULL!" << endl;
                    return TablePtr();
            }            

            query_processing::column_processing::cpu::LogicalQueryPlanPtr log_plan = query_processing::createColumnPlanforDisjunction(table, disjunction, CoGaDB::RuntimeConfiguration::instance().getGlobalDeviceConstraint());
            log_plan->print();
        query_processing::column_processing::cpu::PhysicalQueryPlanPtr phy_plan;
        if (RuntimeConfiguration::instance().isQueryChoppingEnabled()) {
            phy_plan = log_plan->runChoppedPlan();
        } else {
            phy_plan = log_plan->convertToPhysicalQueryPlan();
            phy_plan->print();
            phy_plan->run();
        }
            

            ColumnPtr col = phy_plan->getResult();
            cout << "Physical TOPPO Plan: " << endl;
            phy_plan->print(); 
            PositionListPtr tids = getPositonListfromLookupArray(col);
            if(!tids) return TablePtr(); 
             
            TablePtr result_table = createResultTable(table, tids, mat_stat, "selection");
             
            return result_table ;
        }

        const TablePtr BaseTable::selection(TablePtr table, const KNF_Selection_Expression& knf_expr, MaterializationStatus mat_stat, ParallelizationMode comp_mode){

            if(!table){
                    cout << "Fatal Error: in BaseTable::selection(): input table pointer is NULL!" << endl;
                    return TablePtr();
            }
            if(knf_expr.disjunctions.empty()) return TablePtr();
            //stores the result for each disjunction
            std::vector<PositionListPtr> disjunctions_result_tid_lists(knf_expr.disjunctions.size());
            for(unsigned int i=0;i<knf_expr.disjunctions.size();i++){
                //stores the tid list for each predicate
                std::vector<PositionListPtr> predicate_result_tid_lists(knf_expr.disjunctions[i].size());
                for(unsigned j=0;j<knf_expr.disjunctions[i].size();j++){
                    if(knf_expr.disjunctions[i][j].getPredicateType()==ValueValuePredicate){
                        ColumnPtr col = table->getColumnbyName(knf_expr.disjunctions[i][j].getColumn1Name());
                        if(!col){ 
                                cout << "Error! in BaseTable::selection(): Could not find Column " << knf_expr.disjunctions[i][j].getColumn1Name() << " in Table " << table->getName() << endl; 
                                cout << "In File " << __FILE__ << ":" << __LINE__ << endl;
                                return TablePtr();
                        }
                        ColumnPtr col2= table->getColumnbyName(knf_expr.disjunctions[i][j].getColumn2Name());
                        if(!col2){ 
                                cout << "Error! in BaseTable::selection(): Could not find Column " << knf_expr.disjunctions[i][j].getColumn2Name() << " in Table " << table->getName() << endl; 
                                cout << "In File " << __FILE__ << ":" << __LINE__ << endl;                               
                                return TablePtr();
                        }
                        predicate_result_tid_lists[j]=col->selection(col2,knf_expr.disjunctions[i][j].getValueComparator());
                    }else if(knf_expr.disjunctions[i][j].getPredicateType()==ValueConstantPredicate){
                        ColumnPtr col = table->getColumnbyName(knf_expr.disjunctions[i][j].getColumn1Name());
                        if(!col){ 
                                cout << "Error! in BaseTable::selection(): Could not find Column " << knf_expr.disjunctions[i][j].getColumn1Name() << " in Table " << table->getName() << endl; 
                                cout << "In File " << __FILE__ << ":" << __LINE__ << endl;
                                return TablePtr();
                        }
                        predicate_result_tid_lists[j]=col->selection(knf_expr.disjunctions[i][j].getConstant(),knf_expr.disjunctions[i][j].getValueComparator());
                    }else{
                        cerr << "FATAL ERROR! in BaseTable::selection(): Unknown Predicate Type!" << endl;
                        cerr << "In File " << __FILE__ << ":" << __LINE__ << endl;
                        exit(-1);
                    }
                }
                //merge sorted tid lists (compute the disjunction)
                //PositionListPtr result_tids(new PositionList());
            //use first tid list as result tid list, then perform union with all other lists to get the result for this disjunction
            PositionListPtr result_tids(new PositionList(predicate_result_tid_lists[0]->begin(),
                                                         predicate_result_tid_lists[0]->end()));
                for(unsigned int k=1;k<predicate_result_tid_lists.size();++k){
                     PositionListPtr tmp_tids(new PositionList(result_tids->size()+predicate_result_tid_lists[k]->size()));
                     PositionList::iterator it;
                     it=std::set_union(result_tids->begin(),result_tids->end(),
                     predicate_result_tid_lists[k]->begin(),
                     predicate_result_tid_lists[k]->end(),tmp_tids->begin());
                     tmp_tids->resize(it-tmp_tids->begin());
                     result_tids=tmp_tids;
                }
                disjunctions_result_tid_lists[i]=result_tids;
                if(!quiet && verbose && debug){
                    cout << "Disjunction " << i << "result:" << endl;
                    for(unsigned int i=0;i<result_tids->size();i++){
                        cout << (*result_tids)[i] << endl;
                    }
                }
                
            }
            //intersect tid list for each disjunction to compute final result
            //use first tid list as result tid list, then intersect with all other lists to get the overall result
            PositionListPtr tids(new PositionList(disjunctions_result_tid_lists[0]->begin(),
                                                  disjunctions_result_tid_lists[0]->end()));
            for(unsigned int i=1;i<knf_expr.disjunctions.size();i++){
                PositionListPtr tmp_tids(new PositionList(tids->size()+disjunctions_result_tid_lists[i]->size()));
                PositionList::iterator it;
                it=std::set_intersection(tids->begin(),tids->end(),
                disjunctions_result_tid_lists[i]->begin(),
                disjunctions_result_tid_lists[i]->end(),tmp_tids->begin());
                tmp_tids->resize(it-tmp_tids->begin());
                //it=std::set_intersection (first, first+5, second, second+5, v.begin());
                                               // 10 20 0  0  0  0  0  0  0  0
                //v.resize(it-v.begin());
                
                tids=tmp_tids;
            }
            if(!quiet && verbose && debug){
                cout << "Konjunction result:" << endl;
                for(unsigned int i=0;i<tids->size();i++){
                    cout << (*tids)[i] << endl;
                }
            }
            //selection is completed, now the result table has to be cosntructred!
            TablePtr result_table = createResultTable(table, tids, mat_stat, "selection");
            return result_table;
            
        }
        
        
	const TablePtr BaseTable::projection(TablePtr table, const std::list<std::string>& columns_to_select, MaterializationStatus mat_stat, const ComputeDevice) {
		if(!table){
			cout << "Fatal Error: in BaseTable::projection(): input table pointer is NULL!" << endl;
			return TablePtr();
		}
	assert(mat_stat==MATERIALIZE);
	Timestamp begin,end; 
		begin=getTimestamp();

		TablePtr result_table;

		//if(mat_stat==MATERIALIZE){
		if(table->isMaterialized()){
		if(!quiet && verbose) cout << "Performing Projection on Materialized Table..." << endl;
		list<string>::const_iterator it;
		vector<ColumnPtr> new_schema;
		const std::vector<ColumnPtr>& columns = table->getColumns();
		for(it=columns_to_select.begin(); it!=columns_to_select.end(); it++){
                    //cout << "Projection: Examine Column " << columns[i]->getName() << endl;
                        for(unsigned int i=0;i<columns.size();i++){		
				if(*it==columns[i]->getName()){
					ColumnPtr new_column = columns[i]->copy();
					new_schema.push_back(new_column);
				}
			}
		}
               
		result_table = TablePtr (new Table(string("projection( ")+table->getName()+" )",new_schema)) ;//tmp_schema));
		//TablePtr result_table(new Table(string(""),new_schema)) ;//tmp_schema));

//		}else if(mat_stat==LOOKUP){
		}else{
			if(!quiet && verbose) cout << "Performing Projection on Lookup Table..." << endl;

		LookupTablePtr input_lookup_table = shared_pointer_namespace::static_pointer_cast<LookupTable>(table);
		assert(input_lookup_table!=NULL);
		const std::vector<LookupColumnPtr>& input_lookup_columns = input_lookup_table->getLookupColumns();
		//create copy of LookupColumns, so the can be passed to the new Lookup Table
		vector<LookupColumnPtr> new_lookup_columns;
		if(!quiet && verbose) cout << "Copying Lookup Columns" << endl;
		for(unsigned int i=0;i<input_lookup_columns.size();i++){
			LookupColumnPtr lcol = input_lookup_columns[i]->copy();
			new_lookup_columns.push_back(lcol);
		}
		//collect LookupArrays of new LookupColumns
		std::vector<ColumnPtr> new_lookup_arrays;
		for(unsigned int i=0;i<new_lookup_columns.size();i++){
			ColumnVectorPtr v = new_lookup_columns[i]->getLookupArrays();
			new_lookup_arrays.insert(new_lookup_arrays.end(),v->begin(),v->end());
		}	
                //collect dense Value Columns
                const ColumnVector& dense_value_arrays=input_lookup_table->getDenseValueColumns();
                new_lookup_arrays.insert(new_lookup_arrays.end(),dense_value_arrays.begin(),dense_value_arrays.end());

		list<string>::const_iterator it;
		vector<ColumnPtr> result_lookup_arrays;
		TableSchema new_schema;
                for(it=columns_to_select.begin(); it!=columns_to_select.end(); it++){
                        for(unsigned int i=0;i<new_lookup_arrays.size();i++){
			if(!quiet && verbose) cout << "Projection: Examining Column: " << new_lookup_arrays[i]->getName() << endl;
				if(*it==new_lookup_arrays[i]->getName()){
					new_schema.push_back( Attribut(new_lookup_arrays[i]->getType(), new_lookup_arrays[i]->getName()) ); //create entry in schema
					ColumnPtr new_column = new_lookup_arrays[i]; //lookup Array is already a copy!     //columns[i]->copy();
					result_lookup_arrays.push_back(new_column);
				}
                        }
		}                
                
			result_table = TablePtr(new LookupTable(string("projection( ")+table->getName()+" )",new_schema, new_lookup_columns, result_lookup_arrays) );

		}
		


		end=getTimestamp();
		assert(end>=begin);
		if(print_time_measurement) cout << "Time for Projection: " << end-begin << "ns (" << double(end-begin)/1000000 << "ms)" << endl;

		return result_table;
	}

	const TablePtr BaseTable::join(TablePtr table1, const std::string& join_column_table1, TablePtr table2, const std::string& join_column_table2, 
											 JoinAlgorithm join_alg, MaterializationStatus mat_stat, const ComputeDevice comp_dev){

		if(!table1){
			cout << "Fatal Error: in BaseTable::join(): input table pointer (table1) is NULL!" << endl;
			return TablePtr();
		}	
		if(!table2){
			cout << "Fatal Error: in BaseTable::join(): input table pointer (table2) is NULL!" << endl;
			return TablePtr();
		}			
												 
		//assert(comp_dev!=GPU);

		Timestamp begin,end; 
		begin=getTimestamp();
		//shared_pointer_namespace::shared_ptr<Table> table = shared_pointer_namespace::static_pointer_cast<Table>(table_);

		ColumnPtr col1 = table1->getColumnbyName(join_column_table1);
		if(col1==NULL){
			cout << "table " << table1->getName() << " has no Column named " << join_column_table1 << endl;
			cout << "join failed!" << endl;
			return TablePtr();
		}
		ColumnPtr col2 = table2->getColumnbyName(join_column_table2);
		if(col2==NULL){
			cout << "table " << table2->getName() << " has no Column named " << join_column_table2 << endl;
			cout << "join failed!" << endl;
			return TablePtr();
		}

		//PositionListPairPtr join_tids;
		PositionListPairPtr join_tids( new PositionListPair());
//ofstream file;
//file.open("eval.txt", std::fstream::app);
//Timestamp t;

		if(comp_dev==CPU){
		if(join_alg==SORT_MERGE_JOIN){//SORT_MERGE_JOIN,NESTED_LOOP_JOIN,HASH_JOIN
			if(!quiet)cout << "Perform Sort Merge Join..." << endl;
		join_tids = col1->sort_merge_join(col2);
		}else if(join_alg==NESTED_LOOP_JOIN){
			if(!quiet)cout << "Perform Nested Loop Join..." << endl;
		join_tids = col1->nested_loop_join(col2);
		}else if(join_alg==HASH_JOIN){
			if(!quiet)cout << "Perform Hash Join..." << endl;
//t = getTimestamp();
			if(col1->size()<col2->size()){
				join_tids = col1->hash_join(col2);
			}else{
				join_tids = col2->hash_join(col1); //hash table is build for smaller relation, which is the object the method is called on
				std::swap(join_tids->first,join_tids->second);
			}	
                        /*
			if(col1->size()<col2->size()){
                            cout << "Sizes of Input Columns: col1: " << col1->size() << " col2: " << col2->size() << endl; 
 				join_tids = col1->parallel_hash_join(col2,4);
                                PositionListPairPtr join_tids_reference =  col1->hash_join(col2);
                                if(join_tids_reference->first->size()!=join_tids->first->size()){
                                    cout << "SHJ: size " << join_tids_reference->first->size() << endl;
                                    cout << "PHJ: size " << join_tids->first->size() << endl;  
                                    COGADB_FATAL_ERROR("Parallel Hash Join returned wrong result!","");
                                }
                                bool wrong_result=false;
                                for(unsigned int i=0;i<join_tids_reference->first->size();++i){
                                    if(join_tids_reference->first->at(i)!=join_tids->first->at(i)){
                                        wrong_result=true;
                                        cout << "ERROR PK_TIDS at Position:" << i << " SHJ: " << join_tids_reference->first->at(i) << " PHJ: " << join_tids->first->at(i) << endl;
                                    }
                                   if(join_tids_reference->second->at(i)!=join_tids->second->at(i)){
                                        wrong_result=true;
                                        cout << "ERROR FK_TIDS at Position:" << i << " SHJ: " << join_tids_reference->second->at(i) << " PHJ: " << join_tids->second->at(i) << endl;
                                    }
                                }
                                if(wrong_result){
                                    COGADB_FATAL_ERROR("Parallel Hash Join returned wrong result!","");
                                }
			}else{
				join_tids = col2->parallel_hash_join(col1,4); //hash table is build for smaller relation, which is the object the method is called on
				std::swap(join_tids->first,join_tids->second);
                                PositionListPairPtr join_tids_reference =  col2->hash_join(col1);
                                std::swap(join_tids_reference->first,join_tids_reference->second);
                                if(join_tids_reference->first->size()!=join_tids->first->size()){
                                    cout << "SHJ: size " << join_tids_reference->first->size() << endl;
                                    cout << "PHJ: size " << join_tids->first->size() << endl;  
                                    COGADB_FATAL_ERROR("Parallel Hash Join returned wrong result!","");
                                }
                                bool wrong_result=false;
                                for(unsigned int i=0;i<join_tids_reference->first->size();++i){
                                    if(join_tids_reference->first->at(i)!=join_tids->first->at(i)){
                                        wrong_result=true;
                                        cout << "ERROR PK_TIDS at Position:" << i << " SHJ: " << join_tids_reference->first->at(i) << " PHJ: " << join_tids->first->at(i) << endl;
                                    }
                                   if(join_tids_reference->second->at(i)!=join_tids->second->at(i)){
                                        wrong_result=true;
                                        cout << "ERROR FK_TIDS at Position:" << i << " SHJ: " << join_tids_reference->second->at(i) << " PHJ: " << join_tids->second->at(i) << endl;
                                    }
                                }
                                if(wrong_result){
                                    COGADB_FATAL_ERROR("Parallel Hash Join returned wrong result!","");
                                }
			}//*/


//file << (double) (getTimestamp() - t) / 1000000 << "\t";
		}else{
		cout << "Fatal Error: BaseTable::join(): Unknown Join algorithm!" << endl;
		exit(-1);
		}
		}else if(comp_dev==GPU){
			//cout << "JOIN:" << endl;
			//cout << "LEFT Table has " << table1->getNumberofRows() << "rows "<< endl;
			//cout << "RIGHT Table has " << table2->getNumberofRows() << "rows "<< endl;
			if(table1->getNumberofRows()==0 || table2->getNumberofRows()==0){
				//it is clear taht the result is empty, so we take a shortcut:
				join_tids->first = PositionListPtr( new PositionList() );
				join_tids->second = PositionListPtr( new PositionList() );
			}else{

				//for gpu warmup:
				if (join_alg == SORT_MERGE_JOIN_2)
				{
					gpu::GPU_Base_ColumnPtr primary_column = GPU_Column_Cache::instance().getGPUColumn(col1); //primary host to device
					gpu::GPU_Base_ColumnPtr foreign_column = GPU_Column_Cache::instance().getGPUColumn(col2); //foreign host to device
					std::pair<gpu::GPU_PositionlistPtr,gpu::GPU_PositionlistPtr> gpu_posList = gpu::GPU_Operators::sort_merge_join(primary_column, foreign_column);	
					join_tids->first = copy_PositionList_device_to_host(gpu_posList.first);
					join_tids->second = copy_PositionList_device_to_host(gpu_posList.second);
				}
				else
				{
	//t = getTimestamp();
					gpu::GPU_Base_ColumnPtr primary_column = GPU_Column_Cache::instance().getGPUColumn(col1); //copy primary columns to device
					gpu::GPU_Base_ColumnPtr foreign_column = GPU_Column_Cache::instance().getGPUColumn(col2); //copy foreign columns to device
	//file << (double) (getTimestamp() - t) / 1000000 << "\t";

	//t = getTimestamp();
					std::pair<gpu::GPU_PositionlistPtr,gpu::GPU_PositionlistPtr> gpu_posList = gpu::GPU_Operators::sort_merge_join(primary_column, foreign_column);	//perform join operation on GPU	
	//file << (double) (getTimestamp() - t) / 1000000 << "\t";
				
	//t = getTimestamp();
					join_tids->first = copy_PositionList_device_to_host(gpu_posList.first); //copy primary tids to host memory
					join_tids->second = copy_PositionList_device_to_host(gpu_posList.second); //copy foreign tids to host memory
	//file << (double) (getTimestamp() - t) / 1000000 << "\t";
				}			
			}
		}

		//file.close();


		assert(join_tids!=NULL);

		TableSchema result_schema;
		result_schema=mergeTableSchemas(table1->getSchema(),join_column_table1,table2->getSchema(),join_column_table2);

		assert(join_tids->first->size()==join_tids->second->size());

		TablePtr result_table;

//materialize Table
		if(mat_stat==MATERIALIZE){
//		TablePtr result_table(new Table(string("join( ")+table1->getName()+string(".")+join_column_table1
//																	   +string(",")+table2->getName()+string(".")+join_column_table2+" )",result_schema));

			result_table = TablePtr(new Table(string("join( ")+table1->getName()+string(".")+join_column_table1
											         +string(",")+table2->getName()+string(".")+join_column_table2+" )",result_schema));

			TID index_table1,index_table2;
			for(unsigned int i=0;i<join_tids->first->size();i++){
				index_table1 = (*(join_tids->first))[i];
				index_table2 =  (*(join_tids->second))[i];

				Tuple t;
				const std::vector<ColumnPtr>& columns_ = table1->getColumns();
				for(unsigned int i=0;i<columns_.size();i++)
					t.push_back( columns_[i]->get(index_table1) );

				const std::vector<ColumnPtr>& tab_columns_ = table2->getColumns();
				for(unsigned int i=0;i<tab_columns_.size();i++)
					t.push_back( tab_columns_[i]->get(index_table2) );				

				result_table->insert(t);				
			}

//			return result_table;

		}else if(mat_stat==LOOKUP){

		if(!quiet) cout << "create Lookup Table for resulting Table..." << endl;


			string result_lookup_table_name = string("Lookup_Table_join( ")+table1->getName()+string(".")+join_column_table1
														  +string(",")+table2->getName()+string(".")+join_column_table2+" )";

			//consider join as "two" semi joins
			LookupTablePtr lookup_table_for_table1 = createLookupTableforUnaryOperation(string("AGGREGATED( ")+table1->getName()+" )",table1, join_tids->first);
			LookupTablePtr lookup_table_for_table2 = createLookupTableforUnaryOperation(string("AGGREGATED( ")+table2->getName()+" )",table2, join_tids->second);
			assert(lookup_table_for_table1!=NULL);
			assert(lookup_table_for_table2!=NULL);

			result_table = LookupTable::concatenate(result_lookup_table_name,*lookup_table_for_table1,*lookup_table_for_table2);

	
		}else{
			cout << "Error! Unkown Materialization Option!" << endl;
			return TablePtr(); ////error, return NULL pointer
		}

		end=getTimestamp();
		assert(end>=begin);
		if(print_time_measurement) cout << "Time for Join: " << end-begin << "ns (" << double(end-begin)/1000000 << "ms)" << endl;
		return result_table;
	}
        
        const TablePtr BaseTable::crossjoin(TablePtr table1, TablePtr table2, MaterializationStatus mat_stat){
        
            if(!table1){
                    cout << "Fatal Error: in BaseTable::crossjoin(): input table pointer (table1) is NULL!" << endl;
                    return TablePtr();
            }	
            if(!table2){
                    cout << "Fatal Error: in BaseTable::crossjoin(): input table pointer (table2) is NULL!" << endl;
                    return TablePtr();
            }			
            Timestamp begin,end; 
            begin=getTimestamp();
            
            PositionListPairPtr join_tids( new PositionListPair());
            assert(join_tids!=NULL);
            join_tids->first = PositionListPtr( new PositionList() );
            join_tids->second = PositionListPtr( new PositionList() );
            //perform cross join by generating tids lists
            unsigned int join_column1_size=table1->getNumberofRows();
            unsigned int join_column2_size=table2->getNumberofRows();	
            for(unsigned int i=0;i<join_column1_size;i++){
                for(unsigned int j=0;j<join_column2_size;j++){
                    join_tids->first->push_back(i);
                    join_tids->second->push_back(j);
                }
            }
            //check whether tid lsits are of same length
            assert(join_tids->first->size()==join_tids->second->size());
            //create schema of new table
            TableSchema result_schema;
            result_schema=mergeTableSchemas(table1->getSchema(),"",table2->getSchema(),"");

            TablePtr result_table;

            //materialize Table
            if(mat_stat==MATERIALIZE){
                result_table = TablePtr(new Table(string("crossjoin( ")+table1->getName()+string(",")+table2->getName()+" )",result_schema));
        cout << "HERE COMES THE SORT" << endl;

                TID index_table1,index_table2;
                for(unsigned int i=0;i<join_tids->first->size();i++){
                        index_table1 = (*(join_tids->first))[i];
                        index_table2 =  (*(join_tids->second))[i];
                        Tuple t;
                        const std::vector<ColumnPtr>& columns_ = table1->getColumns();
                        for(unsigned int i=0;i<columns_.size();i++)
                                t.push_back( columns_[i]->get(index_table1) );

                        const std::vector<ColumnPtr>& tab_columns_ = table2->getColumns();
                        for(unsigned int i=0;i<tab_columns_.size();i++)
                                t.push_back( tab_columns_[i]->get(index_table2) );				

                        result_table->insert(t);				
                }
            }else if(mat_stat==LOOKUP){
                if(!quiet) cout << "create Lookup Table for resulting Table..." << endl;
                string result_lookup_table_name = string("Lookup_Table_crossjoin( ")+table1->getName()+string(",")+table2->getName()+" )";
                //consider join as "two" semi joins
                LookupTablePtr lookup_table_for_table1 = createLookupTableforUnaryOperation(string("AGGREGATED( ")+table1->getName()+" )",table1, join_tids->first);
                LookupTablePtr lookup_table_for_table2 = createLookupTableforUnaryOperation(string("AGGREGATED( ")+table2->getName()+" )",table2, join_tids->second);
                assert(lookup_table_for_table1!=NULL);
                assert(lookup_table_for_table2!=NULL);
                //concatenate LookupTables
                result_table = LookupTable::concatenate(result_lookup_table_name,*lookup_table_for_table1,*lookup_table_for_table2);
            }else{
                cout << "Error! Unkown Materialization Option!" << endl;
                return TablePtr(); ////error, return NULL pointer
            }
            end=getTimestamp();
            assert(end>=begin);
            if(print_time_measurement) cout << "Time for CrossJoin: " << end-begin << "ns (" << double(end-begin)/1000000 << "ms)" << endl;
            return result_table;
        }    

	const TablePtr BaseTable::sort(TablePtr table, const std::string& column_name, SortOrder order, MaterializationStatus mat_stat, ComputeDevice comp_dev){
		if(!table){
			cout << "Fatal Error: in BaseTable::sort(): input table pointer is NULL!" << endl;
			return TablePtr();
		}					
		Timestamp begin,end; 
		begin=getTimestamp();
		
		ColumnPtr column = table->getColumnbyName(column_name);
			if(column.get()==NULL){
				cout << "Error! Could not look up Column: " << column_name << endl;
				exit(-1);
			}
			

			PositionListPtr ids;

			if(comp_dev==CPU){
				ids = column->sort(order);
			}else if(comp_dev==GPU){
				//copy original column to GPU RAM
				gpu::GPU_Base_ColumnPtr device_column = GPU_Column_Cache::instance().getGPUColumn(column); //copy_column_host_to_device(*this);
				//compute result tid list on GPU
				gpu::GPU_PositionlistPtr device_tids = gpu::GPU_Operators::sort(device_column, order); 
				if(!device_tids ){
					cout << "SORT Operator on GPU failed!" << endl;
					cout << "For column: '" << column_name << "' with " << column->size() << " elements" << endl;
					return TablePtr (); //return NULL Pointer
				}
				//copy result tid list back to CPU RAM
				ids = copy_PositionList_device_to_host(device_tids);
				if(!ids){
					cout << "COPY Operation from GPU to CPU RAM failed!" << endl;	
					cout << "For column: '" << column_name << "' with " << column->size() << " elements" << endl;
					return TablePtr (); //return NULL Pointer
				}
			}

			assert(ids!=NULL);
			assert(column->size()==ids->size()); //number of input values is equal to number of output values

                        TablePtr result_table = createResultTable(table, ids, mat_stat, "sort");
                        
//		TablePtr result_table;
//		if(mat_stat==MATERIALIZE){
//			
//			 result_table = TablePtr(new Table(string("sort( ")+table->getName()+string(".")+column_name+" )",table->getSchema()));
//			//cout << "adding tuples to new table " << endl;
//			for(unsigned int i=0;i<column->size();i++){
//				TID j = (*ids)[i];
////				cout << "size " << column->size() << "   id: " << j << endl;
////				cout << "Tuple: " << table->fetchTuple(j) << endl;
//				result_table->insert( table->fetchTuple(j) );
//			}
//
//                         
                         
//		}else if(mat_stat==LOOKUP){
//
//			if(!quiet) cout << "create Lookup Table for Table " << table->getName() << " ..." << endl;
//			string name=string("Lookup_Table_sort( ")+table->getName()+string(".")+string(column_name) +string(" )");
//			result_table = createLookupTableforUnaryOperation(name, table, ids);
//
//		}else{
//			cout << "Error! Unkown Materialization Option!" << endl;
//			return TablePtr(); ////error, return NULL pointer
//		}

		end=getTimestamp();
		assert(end>=begin);
		if(print_time_measurement) cout << "Time for Sorting: " << end-begin << "ns (" << double(end-begin)/1000000 << "ms)" << endl;
		return result_table;
	}

       const TablePtr BaseTable::sort(TablePtr table, const std::list<std::string>& column_names, SortOrder order, MaterializationStatus mat_stat, ComputeDevice comp_dev){
  		if(!table){
			cout << "Fatal Error: in BaseTable::sort(): input table pointer is NULL!" << endl;
			return TablePtr();
		}	
                //sort(A1,..,An)
                if(!quiet && verbose){
                    cout << "Sorting after: " ;
                    std::list<std::string>::const_iterator cit;
                    for(cit=column_names.begin();cit!=column_names.end();++cit){
                        cout << *cit << "," << endl;
                    }                
                }
                if(column_names.empty()) return table;
                //sort(An)
                if(column_names.size()==1){
                    return BaseTable::sort(table, column_names.front(), order, mat_stat, comp_dev);
                }
                
                std::list<std::string> tmp_column_names (column_names);
                tmp_column_names.pop_front();
                //sort(Ai+1,..,An))
                table=sort(table, tmp_column_names, order, mat_stat, comp_dev);
                //sort(Ai,..,An))       
                return BaseTable::sort(table, column_names.front(), order, mat_stat, comp_dev);
                
        }
        
        
 

	
	const TablePtr BaseTable::groupby(TablePtr table, const std::string& grouping_column, const std::string& aggregation_column, const std::string& result_column_name, AggregationMethod agg_meth, ComputeDevice comp_dev){
		if(!table){
			cout << "Fatal Error: in BaseTable::groupby(): input table pointer is NULL!" << endl;
			return TablePtr();
		}			
															 
		ColumnPtr key_col = table->getColumnbyName(grouping_column);
		if(key_col==NULL){
			cout << "Error! Could not look up Column: " << grouping_column << endl;
                        cout << "In File " << __FILE__ << ":" << __LINE__ << endl;
                        table->print();
			exit(-1);
		}
		ColumnPtr value_col = table->getColumnbyName(aggregation_column);
		if(value_col==NULL){
			cout << "Error! Could not look up Column: " << aggregation_column << endl;
                        cout << "In File " << __FILE__ << ":" << __LINE__ << endl;
                        table->print();
			exit(-1);
		}

			ColumnPtr result_key_col;			
			ColumnPtr result_value_col; 

			if(comp_dev==CPU){
				//tids=column->selection(value_for_comparison, comp, CPU);

				std::pair<ColumnPtr,ColumnPtr> result_;
                                if(value_col->getType()==INT){
                                    result_= reduce_by_keys(shared_pointer_namespace::static_pointer_cast<ColumnBaseTyped<int> > (key_col), 
                                                           shared_pointer_namespace::static_pointer_cast<ColumnBaseTyped<int> > (value_col),
                                                           agg_meth);
                                }else if(value_col->getType()==FLOAT){
                                    result_= reduce_by_keys(shared_pointer_namespace::static_pointer_cast<ColumnBaseTyped<int> > (key_col), 
                                                           shared_pointer_namespace::static_pointer_cast<ColumnBaseTyped<float> > (value_col),
                                                           agg_meth);                                    
                                }                                              

				result_key_col = result_.first;
				result_value_col = result_.second;
				
			}else if(comp_dev==GPU){
				//copy original column to GPU RAM
				gpu::GPU_Base_ColumnPtr dev_key_col   = GPU_Column_Cache::instance().getGPUColumn(key_col);
				gpu::GPU_Base_ColumnPtr dev_value_col = GPU_Column_Cache::instance().getGPUColumn(value_col);
	
				assert(dev_key_col!=NULL);
				assert(dev_value_col!=NULL);
	
				std::pair<gpu::GPU_Base_ColumnPtr,gpu::GPU_Base_ColumnPtr> result = gpu::GPU_Operators::groupby(dev_key_col, dev_value_col,agg_meth);	

				assert(result.first  != NULL);
				assert(result.second != NULL);
				
				result_key_col = gpu::copy_column_device_to_host(result.first);
				result_value_col = gpu::copy_column_device_to_host(result.second);
			}				

//			TableSchema schema;
//			schema.push_back( Attribut(VARCHAR,grouping_column) );
//			schema.push_back( Attribut(INT,aggregation_column) );
//			TablePtr result_table(new Table(string("groupby( ")+table->getName()+string(".")+grouping_column+" )",schema));
//		


				assert(result_key_col  != NULL);
				assert(result_value_col != NULL);
                        result_value_col->setName(result_column_name);                
			std::vector<ColumnPtr> columns;
			columns.push_back(result_key_col);
			columns.push_back(result_value_col);
			TablePtr result_table(new Table(string("groupby( ")+table->getName()+string(".")+grouping_column+" )",columns));

		return result_table;
	}

        int compareTupleValues(vector<ColumnPtr>& grouping_columns, int j, int k){
            for(unsigned int i=0;i<grouping_columns.size();++i){
                if(!CoGaDB::quiet && CoGaDB::verbose && CoGaDB::debug) cout << "Compare Attribute " << grouping_columns[i]->getName(); // << endl;
                int ret = grouping_columns[i]->compareValuesAtIndexes(j,k); 
                if(!CoGaDB::quiet && CoGaDB::verbose && CoGaDB::debug) cout << " Result: " << ret << endl;
                if(ret) return ret;
            }
            if(!CoGaDB::quiet && CoGaDB::verbose && CoGaDB::debug)  cout << "Tuples are Equal and belong to the same group!" << endl;
            return 0;
        }
        
        vector<ColumnPtr> finalizeGroupingColumns(vector<ColumnPtr> grouping_column_ptrs, PositionListPtr position_list){
            vector<ColumnPtr> distinct_groups; //(grouping_column_ptrs.size());
                    
            for(unsigned int i=0;i<grouping_column_ptrs.size();++i){
                //LookupArray(grouping_column_ptrs[i]->getName(), grouping_column_ptrs[i]->getType(), grouping_column_ptrs[i], position_list); 
                ColumnPtr col = grouping_column_ptrs[i]->gather(position_list);
                distinct_groups.push_back(col);
            }
            return distinct_groups;
        }
        
        
        const TablePtr BaseTable::groupby(TablePtr table, const std::list<std::string>& grouping_columns, std::list<std::pair<string,AggregationMethod> > aggregation_functions, ComputeDevice comp_dev){
            if(!table){
                    cout << "Fatal Error: in BaseTable::groupby(): input table pointer is NULL!" << endl;
                    return TablePtr();

            }
            if(table->getNumberofRows()==0){
                //create empty table of the expected schema
                    TableSchema schema;
                 {    
                    std::list<std::string>::const_iterator cit;
                    vector<ColumnPtr> grouping_column_ptrs;
                    for(cit=grouping_columns.begin();cit!=grouping_columns.end();++cit){
                        ColumnPtr col = table->getColumnbyName(*cit);
                        assert(col!=NULL);
                        schema.push_back(Attribut(col->getType(),col->getName()));
                        //grouping_column_ptrs.push_back(col);
                    }
                 } 
                std::list<std::pair<string,AggregationMethod> >::const_iterator cit;
                for(cit=aggregation_functions.begin();cit!=aggregation_functions.end();++cit){
                    ColumnPtr col = table->getColumnbyName(cit->first);
                    assert(col!=NULL);
                    schema.push_back(Attribut(col->getType(),col->getName()));
                }
                TablePtr empty_table(new Table(string("groupby(")+table->getName()+")",schema)); 
                return empty_table;
            }
            //create grouping key column
            vector<ColumnPtr> grouping_column_ptrs;
            std::list<std::string>::const_iterator cit;
            for(cit=grouping_columns.begin();cit!=grouping_columns.end();++cit){
                if(!CoGaDB::quiet && CoGaDB::verbose && CoGaDB::debug) cout << *cit << ",";
                ColumnPtr col = table->getColumnbyName(*cit);
                assert(col!=NULL);
                grouping_column_ptrs.push_back(col);
            }
            //compareTupleValues(vector<ColumnPtr> grouping_columns, int i, int j)
            std::vector<unsigned int> group_ids(table->getNumberofRows());
            //groupid will always point to the start index where new group starts, hence can later be used to fetch the grouping columns using a left fetch join
            unsigned int groupid=0;
            for(unsigned int i=0;i<table->getNumberofRows()-1;++i){
                if(i==0) group_ids[0]=0;
                //if tuples are unequal, start new group
                int ret=compareTupleValues(grouping_column_ptrs,i,i+1);
                if(!CoGaDB::quiet && CoGaDB::verbose && CoGaDB::debug) cout << "Comp Func: " << ret << endl;
                if(ret){
                //if(compareTupleValues(grouping_column_ptrs,i,i+1)){
                    groupid=i+1;
                }
                //assign current group id
                group_ids[i+1]=groupid;
                //if(i==table->getNumberofRows()-2) group_ids[i+1]=groupid;
            }
            if(!CoGaDB::quiet && CoGaDB::verbose && CoGaDB::debug){
                cout << "Group IDs " << endl;          
                for(unsigned int i=0;i<group_ids.size();++i){
                    cout << group_ids[i] << endl;
                }
            }
            Column<int>* col_typed = new Column<int>("grouping_keys",INT); 
            col_typed->insert(group_ids.begin(),group_ids.end());
            ColumnPtr col(col_typed);
            
            table->addColumn(col); 
            
            vector<ColumnPtr> result_columns;
            {
                std::list<std::pair<string,AggregationMethod> >::const_iterator cit;
                for(cit=aggregation_functions.begin();cit!=aggregation_functions.end();++cit){
                    if(!CoGaDB::quiet && CoGaDB::verbose && CoGaDB::debug) cout << "Group by grouping_keys for column: " << cit->first << endl;
                    TablePtr tmp = BaseTable::groupby(table,std::string("grouping_keys"),cit->first,cit->first,cit->second);
                    ColumnPtr result = tmp->getColumnbyName(cit->first);
                    assert(result!=NULL);
                    
                    if(cit==aggregation_functions.begin()){ 
                        //result_columns.push_back(tmp->getColumnbyName("grouping_keys"));
                       ColumnPtr grouping_column = tmp->getColumnbyName("grouping_keys"); 
                       if(!CoGaDB::quiet && CoGaDB::verbose && CoGaDB::debug) grouping_column->print();
                       assert(grouping_column->getType()==INT);
                       assert(grouping_column->isMaterialized());
                       if(!CoGaDB::quiet && CoGaDB::verbose && CoGaDB::debug) cout << "Type: " << typeid(*grouping_column.get()).name();
                       //exit(0);
                        //shared_pointer_namespace::shared_ptr<Column<int> > int_column = shared_pointer_namespace::dynamic_pointer_cast<Column<int> >( grouping_column);
                        Column<int>* int_column = dynamic_cast<Column<int>* >(grouping_column.get());
                        assert(int_column!=NULL);
                        PositionListPtr tids (new PositionList());
                        for(unsigned int i=0;i<int_column->size();i++){
                            tids->push_back((unsigned int)(*int_column)[i]);
                        }
                        //tids->insert(tids->begin(),int_column->getContent().begin(),int_column->getContent().end());
                        if(!CoGaDB::quiet && CoGaDB::verbose && CoGaDB::debug){
                            cout << "Group TIDS" << endl;
                            for(unsigned int i=0;i<tids->size();i++){
                                cout << "tid: " <<  (*tids)[i] << " original: " << (*int_column)[i]  << endl; 
                            }
                        }
                        vector<ColumnPtr> final_grouping_columns = finalizeGroupingColumns(grouping_column_ptrs,tids);
                        result_columns.insert(result_columns.begin(),final_grouping_columns.begin(),final_grouping_columns.end());
                       
                        
                    }
                    result_columns.push_back(result);
                }
            }
           
            TablePtr result(new Table(string("groupby(")+table->getName()+")",result_columns));
            return result;
        }

        /***************** Aggregation Functions *****************/

        TablePtr BaseTable::ColumnConstantOperation(TablePtr tab, const std::string& col_name, const boost::any& value, const std::string& result_col_name, CoGaDB::ColumnAlgebraOperation operation, const ComputeDevice comp_dev){
        
            if(!tab){
                cout << "Error! In BaseTable::ColumnConstantOperation(): Invalid TablePtr!" << endl;
                cout << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
                return TablePtr();
            }
            //assert(comp_dev==CPU);
            ColumnPtr col = tab->getColumnbyName(col_name);
            if(!col){
                cout << "Error! In BaseTable::ColumnConstantOperation(): Column '" << col_name << "' not found in table '" << tab->getName() << "'" << endl;
                cout << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
                return TablePtr();
            }
            
            ColumnPtr tmp;      
            if(!col->isMaterialized() || col->isCompressed()){ //is col a compressed column or a lookup array?
                tmp = col->materialize();
            }else{ //col is a dense uncompressed array
                tmp = col->copy();
            }
            if(!tmp){
                cerr << "Materialization of Column failed!" << endl;
                cerr << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
                return TablePtr();
            }
	    if(comp_dev==CPU){
		    if(operation==ADD){
		        if(!tmp->add(value)){
		            cerr << "Computation failed!" << endl;
		            cerr << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
		            return TablePtr();
		        }
		    }else if(operation==SUB){
		        if(!tmp->minus(value)){
		            cerr << "Computation failed!" << endl;
		            cerr << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
		            return TablePtr();
		        }              
		    }else if(operation==MUL){
		        if(!tmp->multiply(value)){
		            cerr << "Computation failed!" << endl;
		            cerr << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
		            return TablePtr();
		        }             
		    }else if(operation==DIV){  
		        if(!tmp->division(value)){
		            cerr << "Computation failed!" << endl;
		            cerr << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
		            return TablePtr();
		        }                 
		    }
	    }else if(comp_dev==GPU){
		//copy original column to GPU RAM
		gpu::GPU_Base_ColumnPtr device_column = GPU_Column_Cache::instance().getGPUColumn(tmp);
		if(!device_column){
		   cout << "Error! Host to Device copy operation failed!" << endl;
		   cout << "For Column '" << col->getName() << "' of size '" << col->getSizeinBytes() << "Byte' (" << col->size() << " elements of type " << col->type().name() << ")" << endl;
		   cout << "File: " << __FILE__ << " at Line: " << __LINE__ << endl;
		   printGPUStatus();
		   return TablePtr();
		}

		if(operation==ADD){
			if(!device_column->add(value)){
			    cerr << "Computation failed!" << endl;
			    cerr << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
			    return TablePtr();
			}
		}else if(operation==SUB){
			if(!device_column->minus(value)){
			    cerr << "Computation failed!" << endl;
			    cerr << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
			    return TablePtr();
			}              
		}else if(operation==MUL){
			if(!device_column->multiply(value)){
			    cerr << "Computation failed!" << endl;
			    cerr << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
			    return TablePtr();
			}             
		}else if(operation==DIV){  
			if(!device_column->division(value)){
			    cerr << "Computation failed!" << endl;
			    cerr << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
			    return TablePtr();
			}                 
		}
		//fetch result from GPU
		tmp=gpu::copy_column_device_to_host(device_column);
		if(!GPU_Column_Cache::instance().removeGPUColumn(device_column)){
			    cerr << "Warning: Could not remove column from GPU cache!" << endl;
			    cerr << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
		   	    printGPUStatus();
			    //return TablePtr();
		}

		
		if(!tmp){
		   cout << "Error! GPU Operator 'ColumnConstantOperation' failed!" << endl;
		   cout << "File: " << __FILE__ << " at Line: " << __LINE__ << endl;
		   printGPUStatus();
		   return TablePtr();
		}
		/*
		//copy result tid list back to CPU RAM
		tids = copy_PositionList_device_to_host(device_tids);*/	    
	    }else{
		   cout << "Fatal Error! Unknown Device Type!" << endl;
		   cout << "File: " << __FILE__ << " at Line: " << __LINE__ << endl;
		   printGPUStatus();
		   exit(-1);		
	    }
            tmp->setName(result_col_name);
            TablePtr ret_tab=tab;
  
            if(tab->isMaterialized()){
                //create a LookupTable with same content as original table
                PositionListPtr tids(new PositionList());

                for(unsigned int i=0;i<tab->getNumberofRows();i++){
                    tids->push_back(i);
                }
                LookupTablePtr result_lookup_table = createLookupTableforUnaryOperation(string("lookup(")+tab->getName()+")",tab, tids);
                ret_tab=result_lookup_table;
            }
            //append column to Lookup Table
            ret_tab->addColumn(tmp);
            return ret_tab; //result_lookup_table->addColumn(tmp);
        }
  
        TablePtr BaseTable::ColumnAlgebraOperation(TablePtr tab, const std::string& col1_name, const std::string& col2_name, const std::string& result_col_name, CoGaDB::ColumnAlgebraOperation operation, const ComputeDevice comp_dev){
            if(!tab){
                cout << "Error! In BaseTable::ColumnAlgebraOperation(): Invalid TablePtr!" << endl;
                cout << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
                return TablePtr();
            }
            //assert(comp_dev==CPU);
           
            ColumnPtr col1 = tab->getColumnbyName(col1_name);
            //if(!col1) return false;
            if(!col1){
                cout << "Error! In BaseTable::ColumnConstantOperation(): Column '" << col1_name << "' not found in table '" << tab->getName() << "'" << endl;
                cout << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
                return TablePtr();
            } 
            ColumnPtr col2 = tab->getColumnbyName(col2_name);
            //if(!col2) return false;   
            if(!col2){
                cout << "Error! In BaseTable::ColumnConstantOperation(): Column '" << col2_name << "' not found in table '" << tab->getName() << "'" << endl;
                cout << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
                return TablePtr();
            }             
            
            ColumnPtr tmp;      
            if(!col1->isMaterialized() || col1->isCompressed()){ //is col a compressed column or a lookup array?
                tmp = col1->materialize();
            }else{ //col is a dense uncompressed array
                tmp = col1->copy();
            }
                        
            if(!tmp){
                cerr << "Materialization of Column failed!" << endl;
                cerr << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
                return TablePtr();
            }
            
            if(comp_dev==CPU){
                if(operation==ADD){
                    if(!tmp->add(col2)){
                        cerr << "Computation failed!" << endl;
                        cerr << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
                        return TablePtr();
                    }
                }else if(operation==SUB){
                    if(!tmp->minus(col2)){
                        cerr << "Computation failed!" << endl;
                        cerr << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
                        return TablePtr();
                    }              
                }else if(operation==MUL){
                    if(!tmp->multiply(col2)){
                        cerr << "Computation failed!" << endl;
                        cerr << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
                        return TablePtr();
                    }            
                }else if(operation==DIV){  
                    if(!tmp->division(col2)){
                        cerr << "Computation failed!" << endl;
                        cerr << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
                        return TablePtr();
                    }                 
                }
            }else if(comp_dev==GPU){
                //materialize second column as well, because only materialized columns can be copied to GPU
                ColumnPtr tmp2;      
                if(!col2->isMaterialized() || col2->isCompressed()){ //is col a compressed column or a lookup array?
                      tmp2 = col2->materialize();
                }else{ //col is a dense uncompressed array
                      tmp2 = col2->copy();
                }
                /*copy original columns to GPU RAM*/
                assert(tmp->isMaterialized() && !tmp->isCompressed());
                gpu::GPU_Base_ColumnPtr device_column1 = GPU_Column_Cache::instance().getGPUColumn(tmp);
                if(!device_column1){
                     cout << "Error! Host to Device copy operation failed!" << endl;
                     cout << "For Column '" << tmp->getName() << "' of size '" << tmp->getSizeinBytes() << "Byte' (" << tmp->size() << " elements of type " << tmp->type().name() << ")" << endl;
                     cout << "File: " << __FILE__ << " at Line: " << __LINE__ << endl;
                     printGPUStatus();
                     return TablePtr();
                }      
                gpu::GPU_Base_ColumnPtr device_column2 = GPU_Column_Cache::instance().getGPUColumn(tmp2);
                if(!device_column2){
                     cout << "Error! Host to Device copy operation failed!" << endl;
                     cout << "For Column '" << tmp2->getName() << "' of size '" << tmp2->getSizeinBytes() << "Byte' (" << tmp2->size() << " elements of type " << tmp2->type().name() << ")" << endl;
                     cout << "File: " << __FILE__ << " at Line: " << __LINE__ << endl;
                     printGPUStatus();
                     return TablePtr();
                }                  
                 
                if(!tmp){
                    cerr << "Materialization of Column failed!" << endl;
                    cerr << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
                    return TablePtr();
                }
                /*Check which aggregation function should be executed*/
                if(operation==ADD){
                    if(!device_column1->add(device_column2)){
                        cerr << "Computation failed!" << endl;
                        cerr << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
                        return TablePtr();
                    }
                }else if(operation==SUB){
                    if(!device_column1->minus(device_column2)){
                        cerr << "Computation failed!" << endl;
                        cerr << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
                        return TablePtr();
                    }              
                }else if(operation==MUL){
                    if(!device_column1->multiply(device_column2)){
                        cerr << "Computation failed!" << endl;
                        cerr << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
                        return TablePtr();
                    }            
                }else if(operation==DIV){  
                    if(!device_column1->division(device_column2)){
                        cerr << "Computation failed!" << endl;
                        cerr << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
                        return TablePtr();
                    }                 
                }                
                //fetch result from GPU
		tmp=gpu::copy_column_device_to_host(device_column1);
                //Tell the GPUCache that it can remove cached intermediate results
		if(!GPU_Column_Cache::instance().removeGPUColumn(device_column1)){
			    cerr << "Warning: Could not remove column from GPU cache!" << endl;
			    cerr << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
		   	    printGPUStatus();
			    //return TablePtr();
		}
                if(!GPU_Column_Cache::instance().removeGPUColumn(device_column2)){
			    cerr << "Warning: Could not remove column from GPU cache!" << endl;
			    cerr << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
		   	    printGPUStatus();
			    //return TablePtr();
		}

		if(!tmp){
		   cout << "Error! GPU Operator 'ColumnAlgebraOperation' failed!" << endl;
		   cout << "File: " << __FILE__ << " at Line: " << __LINE__ << endl;
		   printGPUStatus();
		   return TablePtr();
		}
                
            }else{
 		   cout << "Fatal Error! Unknown Device Type!" << endl;
		   cout << "File: " << __FILE__ << " at Line: " << __LINE__ << endl;
		   printGPUStatus();
		   exit(-1);	           
            }
            tmp->setName(result_col_name);
            //return tab->addColumn(tmp);       
            TablePtr ret_tab=tab;
  
            if(tab->isMaterialized()){
                //create a LookupTable with same content as original table
                PositionListPtr tids(new PositionList());

                for(unsigned int i=0;i<tab->getNumberofRows();i++){
                    tids->push_back(i);
                }
                LookupTablePtr result_lookup_table = createLookupTableforUnaryOperation(string("lookup(")+tab->getName()+")",tab, tids);
                ret_tab=result_lookup_table;
            }
            //append column to Lookup Table
            ret_tab->addColumn(tmp);
            return ret_tab; //result_lookup_table->addColumn(tmp);        
        }


      
	/* Utility Functions */

        /*
        bool BaseTable::addColumn(ColumnPtr col){
            if(!col) return true;
            this->
            this->schema_.push_back(Attribut(col->getType(),col->getName()));
            
        }*/
        
        TablePtr BaseTable::AddConstantValueColumnOperation(TablePtr tab, const std::string& col_name, AttributeType type, const boost::any& value, const ComputeDevice comp_dev){
            if(!tab){
                cout << "Error! In BaseTable::AddConstantValueColumnOperation(): Invalid TablePtr!" << endl;
                cout << "In File " << __FILE__ << " Line: " << __LINE__ << endl;
                return TablePtr();
            }       
            assert(comp_dev==CPU);
            ColumnPtr result_column;
            TablePtr ret_tab=tab;
            if(type==INT){
                boost::shared_ptr<Column<int> > col (new Column<int>(col_name,type));
                std::vector<int>& array=col->getContent();
                //add #row times the value to the column (allocates memory and initializes it in one step))
                array.resize(tab->getNumberofRows(),boost::any_cast<int>(value));
                result_column=col;
                //std::fill(array.begin(),array.end(),boost::any_cast<int>(value));
            }else if(type==FLOAT){
                boost::shared_ptr<Column<float> > col (new Column<float>(col_name,type));
                std::vector<float>& array=col->getContent();
                //add #row times the value to the column (allocates memory and initializes it in one step))
                array.resize(tab->getNumberofRows(),boost::any_cast<float>(value)); 
                result_column=col;
            }else if(type==VARCHAR){
                boost::shared_ptr<Column<string> > col (new Column<string>(col_name,type));
                std::vector<string>& array=col->getContent();
                //add #row times the value to the column (allocates memory and initializes it in one step))
                array.resize(tab->getNumberofRows(),boost::any_cast<string>(value));
                result_column=col;
            }else{
                cerr << "Fatal Error: In BaseTable::AddConstantValueColumnOperation(): Invalid Column Type: " << type << endl;
                cerr << "In File " << __FILE__ << " Line " << __LINE__ << endl;
                exit(-1);
            }
            
            if(tab->isMaterialized()){
                //create a LookupTable with same content as original table
                PositionListPtr tids(new PositionList());

                for(unsigned int i=0;i<tab->getNumberofRows();i++){
                    tids->push_back(i);
                }
                LookupTablePtr result_lookup_table = createLookupTableforUnaryOperation(string("lookup(")+tab->getName()+")",tab, tids);
                ret_tab=result_lookup_table;
            } 
            ret_tab->addColumn(result_column);
            return ret_tab;
        }

        
	const LookupTablePtr createLookupTableforUnaryOperation(const std::string& lookup_table_name,const TablePtr table, PositionListPtr ids){
   	assert(table!=NULL);
   	assert(ids!=NULL);	
		LookupTablePtr result_table;
		if(table->isMaterialized()){
			//table is materialized, just create view (Lookup Table) on table (aggregation not neccessary)
			if(!quiet) cout << "passed table is Materialized..." << endl;

			std::vector<LookupColumnPtr> lookup_columns;
			LookupColumnPtr lookup_col ( new LookupColumn(table,ids ) );
			assert(lookup_col!=NULL);

			lookup_columns.push_back(lookup_col);
			ColumnVectorPtr lookup_arrays_lookup_table = lookup_col->getLookupArrays();
			//cout << "Lookup Arrays of result table: " <<  endl;

			result_table = LookupTablePtr(new LookupTable(lookup_table_name,
															table->getSchema(),
															lookup_columns,
															*lookup_arrays_lookup_table));

		}else{
			if(!quiet) cout << "passed table is LookupTable..." << endl;
			//is lookuptable, aggregate results
			LookupTablePtr input_lookup_table = shared_pointer_namespace::static_pointer_cast<LookupTable>(table);
			assert(input_lookup_table!=NULL);
			LookupTablePtr result_lookup_table = LookupTable::aggregate(lookup_table_name, *input_lookup_table, LookupColumn(table,ids ) );
			result_table =  result_lookup_table;
			//no concatenation necessary because this is unary operation
	   }
		return result_table;
	}
}; //end namespace CogaDB

