#include <thrust/version.h>
#include <thrust/device_vector.h>
#include <thrust/iterator/zip_iterator.h>
#include <thrust/partition.h>
#include <thrust/copy.h>
#include <thrust/functional.h>
#include <thrust/sequence.h>
#include <thrust/generate.h>

#include <cstdlib>
#include <cstdio>
#include <iostream>

#include <gpu/gpu_algorithms.hpp>
#include <gpu/gpu_typed_base_column.hpp>
#include <gpu/gpu_positionlist.hpp>
#include <gpu/gpu_functors.hpp>

using namespace std;

namespace CoGaDB{

	namespace gpu{
	
/*
	
template <typename T>
struct Functor_Selection_Tuple_Equal : public thrust::unary_function<thrust::tuple<T,unsigned int>&,bool>
{
  const T a;
  Functor_Selection_Tuple_Equal(T _a) : a(_a) {}

  __host__ __device__
  bool operator()(const thrust::tuple<T,unsigned int>& x) const
  { 
    return ( thrust::get<0>(x)==a);
  }
};

template <typename T>
struct Functor_Selection_Tuple_Lesser : public thrust::unary_function<thrust::tuple<T,unsigned int>&,bool>
{
  const T a;
  
  Functor_Selection_Tuple_Lesser(T _a) : a(_a) {}

  __host__ __device__
  bool operator()(const thrust::tuple<T,unsigned int>& x) const
  { 
    return (thrust::get<0>(x)<a);
  }
};

template <typename T>
struct Functor_Selection_Tuple_Lesser_Equal : public thrust::unary_function<thrust::tuple<T,unsigned int>&,bool>
{
  const T a;
  
  Functor_Selection_Tuple_Lesser_Equal(T _a) : a(_a) {}

  __host__ __device__
  bool operator()(const thrust::tuple<T,unsigned int>& x) const
  { 
    return (thrust::get<0>(x)<=a);
  }
};

template <typename T>
struct Functor_Selection_Tuple_Greater : public thrust::unary_function<thrust::tuple<T,unsigned int>&,bool>
{
  const T a;
  
  Functor_Selection_Tuple_Greater(T _a) : a(_a) {}

  __host__ __device__
  bool operator()(const thrust::tuple<T,unsigned int>& x) const
  { 
    return (thrust::get<0>(x)>a);
  }
};

template <typename T>
struct Functor_Selection_Tuple_Greater_Equal : public thrust::unary_function<thrust::tuple<T,unsigned int>&,bool>
{
  const T a;
  
  Functor_Selection_Tuple_Greater_Equal(T _a) : a(_a) {}

  __host__ __device__
  bool operator()(const thrust::tuple<T,unsigned int>& x) const
  { 
    return (thrust::get<0>(x)>=a);
  }
};

template <typename T>
struct Functor_Tuple_to_TID : public thrust::unary_function<thrust::tuple<T,unsigned int>&,unsigned int>
{
  __host__ __device__
  unsigned int operator()(const thrust::tuple<T,unsigned int>& x) const
  { 
    return thrust::get<1>(x); //return tid
  }
};
*/

template <typename T>
void selection_copy_if_approach(thrust::device_vector<T>& values, T value, ValueComparator comp){

	typedef typename thrust::tuple<typename  thrust::device_vector<T>::iterator,thrust::device_vector<unsigned int>::iterator> Value_TID_Iterator_Tuple;

	thrust::device_vector<unsigned int> tids(values.size()); //reserve memory for tids according to number of values 
	thrust::sequence(tids.begin(),tids.end(), 0); //fill with sequence 0,1,2,3 ...
	//vectors for result data
	thrust::device_vector<T>	result_values(values.size());
	thrust::device_vector<unsigned int>	result_tids(tids.size());
	
	thrust::zip_iterator<Value_TID_Iterator_Tuple> result_first = thrust::make_zip_iterator(thrust::make_tuple(result_values.begin(), result_tids.begin()));
	
	//first and last zip iterator, treat both sequences as one tupel sequence
	thrust::zip_iterator<Value_TID_Iterator_Tuple> first = thrust::make_zip_iterator(thrust::make_tuple(values.begin(), tids.begin()));
	thrust::zip_iterator<Value_TID_Iterator_Tuple> last  = thrust::make_zip_iterator(thrust::make_tuple(values.end(), tids.end()));
	//end_copy iterator will be pointing to the end of the new array
	thrust::zip_iterator<Value_TID_Iterator_Tuple> end_copy;
	//copy only data that satisfies predicate to result destination
	if(comp==EQUAL){
		end_copy = thrust::copy_if(first,last, result_first, Functor_Selection_Tuple_Equal<T>(value));
	}else if(comp==LESSER){
		end_copy = thrust::copy_if(first,last, result_first, Functor_Selection_Tuple_Lesser<T>(value));
	}else if(comp==LESSER_EQUAL){
		end_copy = thrust::copy_if(first,last, result_first, Functor_Selection_Tuple_Lesser_Equal<T>(value));
	}else if(comp==GREATER){
		end_copy = thrust::copy_if(first,last, result_first, Functor_Selection_Tuple_Greater<T>(value));
	}else if(comp==GREATER_EQUAL){
		end_copy = thrust::copy_if(first,last, result_first, Functor_Selection_Tuple_Greater_Equal<T>(value));
	}else{
		std::cout << "FATAL ERROR: Unknown Comparison Type!" << std::endl;
	}
	//modify size, so vector only contains tids which belong to result (end_copy to last don't, its all zeros) 
	result_tids.resize(end_copy-result_first); 

	std::cout << "Result TIDS:" << std::endl;
	thrust::copy(result_tids.begin(), result_tids.end(), std::ostream_iterator<unsigned int>( std::cout,"\n"));
	std::cout << "==============================================" << std::endl;
}


template <typename T>
const GPU_PositionlistPtr selection_stable_partition_approach(thrust::device_vector<T>& values_, T value, ValueComparator comp){

    GPU_PositionlistPtr result;

     try{
	//make a COPY of input vector, since this funtion will change the content of the processed data
	thrust::device_vector<T> values(values_); 

	typedef typename thrust::tuple<typename thrust::device_vector<T>::iterator,thrust::device_vector<unsigned int>::iterator> Value_TID_Iterator_Tuple;

	thrust::device_vector<unsigned int> tids(values.size()); //reserve memory for tids according to number of values 
	thrust::sequence(tids.begin(),tids.end(), 0); //fill with sequence 0,1,2,3 ...
	
	//first and last zip iterator, treat both sequences as one tupel sequence
	thrust::zip_iterator<Value_TID_Iterator_Tuple> first = thrust::make_zip_iterator(thrust::make_tuple(values.begin(), tids.begin()));
	thrust::zip_iterator<Value_TID_Iterator_Tuple> last  = thrust::make_zip_iterator(thrust::make_tuple(values.end(), tids.end()));

	//iterator will point to split point in array
	thrust::zip_iterator<Value_TID_Iterator_Tuple> bound2;
	if(comp==EQUAL){
		bound2 = thrust::stable_partition(first, last, Functor_Selection_Tuple_Equal<T>(value));
	}else if(comp==LESSER){
		bound2 = thrust::stable_partition(first, last, Functor_Selection_Tuple_Lesser<T>(value));
	}else if(comp==LESSER_EQUAL){
		bound2 = thrust::stable_partition(first, last, Functor_Selection_Tuple_Lesser_Equal<T>(value));
	}else if(comp==GREATER){
		bound2 = thrust::stable_partition(first, last, Functor_Selection_Tuple_Greater<T>(value));
	}else if(comp==GREATER_EQUAL){
		bound2 = thrust::stable_partition(first, last, Functor_Selection_Tuple_Greater_Equal<T>(value));	
	}else{
		std::cout << "FATAL ERROR: Unknown Comparison Type!" << std::endl;
	}
	
	thrust::device_vector<unsigned int> result_tids(bound2-first); //reserve space
	thrust::copy(thrust::make_transform_iterator(first,Functor_Tuple_to_TID<unsigned int>()), 
					 thrust::make_transform_iterator(bound2,Functor_Tuple_to_TID<unsigned int>()), 
					 result_tids.begin()) ;
	result=GPU_PositionlistPtr(new Impl_GPU_Positionlist(result_tids));
	
//	std::cout << "Result TIDS:" << std::endl;
//	thrust::copy(result_tids.begin(), result_tids.end(), std::ostream_iterator<unsigned int>( std::cout,"\n"));
//	std::cout << "==============================================" << std::endl;

     }catch(thrust::system::detail::bad_alloc& e){
        cout << e.what() << endl;
	printGPUStatus();
        result=GPU_PositionlistPtr();
     }
	return result;

}

	
	
	
//			template <typename T>
//			const thrust::device_vector<TID>* gpu_selection_tids(const thrust::device_vector<T>& column, T value, ValueComparator comp){
//	
//				typedef thrust::device_vector<thrust::pair<T,unsigned int> > device_vector;
//				typename device_vector::iterator bound2;

//				if(comp==EQUAL){
//					//bound2 = thrust::stable_partition(dev_col.begin(), dev_col.end(), select_equal_pair_functor<T>(value));
//				}else if(comp==LESSER){
//					//bound2 = thrust::stable_partition(dev_col.begin(), dev_col.end(), select_lesser_pair_functor<T>(value));
//				}else if(comp==GREATER){
//					//bound2 = thrust::stable_partition(dev_col.begin(), dev_col.end(), select_greater_pair_functor<T>(value));
//				}else{
//					std::cout << "FATAL Error! GPU selection: invalid ValueComparator: " << comp << std::endl;
//				}
//	
//				return NULL;
//			}
	
			
			const GPU_PositionlistPtr GPU_Operators::selection(GPU_Base_ColumnPtr column, const boost::any& value_for_comparison, const ValueComparator comp){
				if(!column)
					return GPU_PositionlistPtr(); //return null pointer
			        GPU_PositionlistPtr result;
				if(column->getType()==INT){
					//std::cout << "perform selection on GPU for type INT" << std::endl;
//					GPU_Typed_Base_Column<int>* d_col_ptr = static_cast<GPU_Typed_Base_Column<int>*>(column);
					//shared_ptr<GPU_Typed_Base_Column<int> > d_col_ptr = static_cast<GPU_Typed_Base_Column<int>*>(column);					
					shared_pointer_namespace::shared_ptr<GPU_Typed_Base_Column<int> > d_col_ptr = shared_pointer_namespace::static_pointer_cast<GPU_Typed_Base_Column<int> >(column); 	

					if(!d_col_ptr)	
						return GPU_PositionlistPtr(); //return null pointer
						

					result = selection_stable_partition_approach(d_col_ptr->getContent(), boost::any_cast<int>(value_for_comparison),comp);

				}else if(column->getType()==FLOAT){	
					//std::cout << "perform selection on GPU for type FLOAT" << std::endl;
//					GPU_Typed_Base_Column<float>* d_col_ptr = static_cast<GPU_Typed_Base_Column<float>*>(column);
					shared_pointer_namespace::shared_ptr<GPU_Typed_Base_Column<float> > d_col_ptr = shared_pointer_namespace::static_pointer_cast<GPU_Typed_Base_Column<float> >(column); 	
				
					if(!d_col_ptr)	
						return GPU_PositionlistPtr(); //return null pointer

				 result = selection_stable_partition_approach(d_col_ptr->getContent(), boost::any_cast<float>(value_for_comparison),comp);
				}				
			
				return result;
			} 
	



	}; //end namespace gpu

}; //end namespace CoGaDB

