#include <thrust/version.h>

//thrust includes
#include <thrust/device_vector.h>
#include <thrust/copy.h>
#include <thrust/sort.h>
#include <thrust/partition.h>
#include <thrust/functional.h>

#include <gpu/gpu_algorithms.hpp>
#include <gpu/gpu_typed_base_column.hpp>
#include <gpu/gpu_positionlist.hpp>
#include <gpu/gpu_functors.hpp>
#include <utility>

#include <util/time_measurement.hpp>
#include <iostream>
#include <fstream>

namespace CoGaDB{

	namespace gpu{
	
			/**
			*	computes the index+1 of the element with the highest index, that is smaller 
			*	or equal to 'elem' in the array 'in' between the indices 'startidx' and 'endIdx'
			*	using binary search
			*/	
			template<typename T> __device__ inline TID binary_index_search(T * in, T elem, long startIdx, long endIdx){
				long from = startIdx;
				long to = endIdx;
				long mid = (long) (from + to) / 2;
				
				while (from <= to && !(in[mid - 1] <= elem && in[mid] > elem))
				{
					if (in[mid] > elem)
						to = mid - 1;
					else
						from = mid + 1;
					mid = (long) (from + to) / 2;
				}
				return (mid == endIdx && in[endIdx] <= elem) ? mid + 1 : mid;
			}

			/**
			*	stores the indices of matching elements of the arrays 'primary' and 'foreign'
			*	into 'result' in between the index-ranges of 'priMin' to 'priMax' for 'primary'
			* 	and 'forMin' to 'forMax'
			*
			*/
			template<typename T> __device__ inline	void merge_ads (T * primary, TID * primary_idx, T * foreign, long priMin, long priMax, long forMin, long forMax, TID * result){
  				long i = priMin; long j = forMin; long k = forMin;

				// iterate to the end of one of the arrays
				while (i < priMax && j < forMax)
  				{
					
					if (primary[i] < foreign[j])
						i++;
					else if (primary[i] == foreign[j++])					
						result[k++] = primary_idx[i];
  				}

				// if there are still foreign-key-values left, fill up  the result with the last index of the primary-key-array
 				while (j++ < forMax)
	  				result[k++] = primary_idx[i];
			}

			/**
			 * merges two columns ("primary" and "foreign") and writes the computed TIDs of the primary column (corresponding to the foreign TIDs) into "result"
			 *
			 * 3 available modes:
			 *			0 - binarysearch AND merge (for inner-thread sync)
			 *			1 - binarysearch only (for hostside block sync)
			 *			2 - merge only (for hostside block sync)
			 */
			template<typename T> __global__	void mergeColumns(T * primary, TID * primary_idx, long primarysize
				, T * foreign, long foreignsize, TID * result, long threadCount, long * bin_temp, long blocksize, int mode){

				TID idx = blockDim.x * blockIdx.x + threadIdx.x;

				long lx = blocksize * idx;
				long rx = lx + blocksize;
	
				//check if in primary's bounds
				if (lx >= primarysize)
					return;
				//set bounds to max if beyond
				rx = (rx > primarysize) ? primarysize : rx;

				if (mode != 2) //search corresponding blocks in foreign
					bin_temp[idx] = idx == threadCount - 1 ? foreignsize : binary_index_search<T>(foreign, primary[rx - 1], 0, foreignsize);

				if (mode == 0) //sync if inner-thread-mode
					__syncthreads();

				if (mode != 1)//merge corresponding blocks
					merge_ads(primary, primary_idx, foreign, lx, rx, idx == 0 ? 0 : bin_temp[idx - 1], bin_temp[idx], result);					
			}

			/**
			*	sort-merge-join implementation
			*
			*/

			template<typename T> 
			const std::pair<GPU_PositionlistPtr,GPU_PositionlistPtr>  my_sort_merge_join(shared_pointer_namespace::shared_ptr<GPU_Typed_Base_Column<T> > t1, 	
				shared_pointer_namespace::shared_ptr<GPU_Typed_Base_Column<T> > t2){

//ofstream file;
//file.open("eval.txt", std::fstream::app);
//Timestamp t;
				//get device vector with join-data (already on gpu-memory)
				thrust::device_vector<T>& primary = t1->getContent();
				thrust::device_vector<T>& foreign = t2->getContent();

				//init vectors for TID-lists
				thrust::device_vector<TID> primary_ids(primary.size());
				thrust::device_vector<TID> foreign_ids(foreign.size());

				//init vectors for TID-lists
				thrust::sequence(primary_ids.begin(), primary_ids.end());
				thrust::sequence(foreign_ids.begin(), foreign_ids.end());
//t = getTimestamp();
				//	sort columns (AND TID-lists)
				thrust::sort_by_key(primary.begin(), primary.end(), primary_ids.begin());
				thrust::sort_by_key(foreign.begin(), foreign.end(), foreign_ids.begin());
//file << (double) (getTimestamp() - t) / 1000000 << "\t";
				//	set output list (primary only)
				thrust::device_vector<TID> primary_ids_result(foreign.size());
				
//Timestamp s2 = getTimestamp();

				//compute number of threads + blocks
				const int minElemPerThread = 8;
				const int maxThreadsPerBlock = 512;
				int nextPowerOfTwo1, nextPowerOfTwo2;
				for (nextPowerOfTwo1 = 1; nextPowerOfTwo1 < primary.size(); nextPowerOfTwo1 <<= 1); 
				for (nextPowerOfTwo2 = 1; nextPowerOfTwo2 < foreign.size(); nextPowerOfTwo2 <<= 1); 

				int numThreads = (nextPowerOfTwo1 > nextPowerOfTwo2) ? nextPowerOfTwo2 : nextPowerOfTwo1;

				numThreads = numThreads > minElemPerThread ? numThreads / minElemPerThread : 1;

				int threadsPerBlock = (numThreads > maxThreadsPerBlock) ? maxThreadsPerBlock : numThreads;
				int blockCount = numThreads > threadsPerBlock ? numThreads/threadsPerBlock : 1;

				//	get raw pointers for custom kernel usage:
				T * d_primary = thrust::raw_pointer_cast(&primary[0]);
				TID * d_primary_idx = thrust::raw_pointer_cast(&primary_ids[0]);
				T * d_foreign = thrust::raw_pointer_cast(&foreign[0]);
				TID* d_primary_idx_result = thrust::raw_pointer_cast(&primary_ids_result[0]);
				thrust::device_vector<long> binary_temp(numThreads);
				long * d_binary_temp = raw_pointer_cast(&binary_temp[0]);
//t = getTimestamp();
				if (blockCount == 1) //use (faster) join within one Thread-Block (uses inner-block sync)
					mergeColumns<<<blockCount, threadsPerBlock>>>(d_primary, d_primary_idx, primary.size(), d_foreign, foreign.size()
						, d_primary_idx_result, numThreads, d_binary_temp, nextPowerOfTwo1 / numThreads, 0);
				else // join with more blocks (uses cudaDeviceSynchronize() for block-sync)
				{
					mergeColumns<<<blockCount, threadsPerBlock>>>(d_primary, d_primary_idx, primary.size(), d_foreign, foreign.size()
						, d_primary_idx_result, numThreads, d_binary_temp, nextPowerOfTwo1 / numThreads, 1);
					cudaDeviceSynchronize();
					mergeColumns<<<blockCount, threadsPerBlock>>>(d_primary, d_primary_idx, primary.size(), d_foreign, foreign.size()
						, d_primary_idx_result, numThreads, d_binary_temp, nextPowerOfTwo1 / numThreads, 2);
				}
//file << (double) (getTimestamp() - t) / 1000000 << "\t";
//file.close("eval.txt");
				std::pair<GPU_PositionlistPtr,GPU_PositionlistPtr> result;
				result.first=GPU_PositionlistPtr(new Impl_GPU_Positionlist(primary_ids_result));
				result.second= GPU_PositionlistPtr(new Impl_GPU_Positionlist(foreign_ids));
				return result;

			}
			
			const std::pair<GPU_PositionlistPtr,GPU_PositionlistPtr>  GPU_Operators::sort_merge_join(GPU_Base_ColumnPtr t1, GPU_Base_ColumnPtr t2){
						//cout << "Perform GPU Join!" << endl;	
				if(!t1)
				return  std::pair<GPU_PositionlistPtr,GPU_PositionlistPtr>();
				if(!t2)
					return std::pair<GPU_PositionlistPtr,GPU_PositionlistPtr>();

				if(t1->type()!=t2->type())
					return std::pair<GPU_PositionlistPtr,GPU_PositionlistPtr>();
					
				if(t1->getType()!=t2->getType())
					return std::pair<GPU_PositionlistPtr,GPU_PositionlistPtr>();						

				if(t1->size()==0 || t2->size()==0)
					return std::pair<GPU_PositionlistPtr,GPU_PositionlistPtr>(GPU_PositionlistPtr(new Impl_GPU_Positionlist()),GPU_PositionlistPtr(new Impl_GPU_Positionlist()));					

				//string name = device_column->getName();
				AttributeType type = t1->getType();
				//cout << "Processing Column " << name << " with type " << type << endl;
				std::pair<GPU_PositionlistPtr,GPU_PositionlistPtr> result_tids;
				if(type==INT){
					shared_pointer_namespace::shared_ptr<GPU_Typed_Base_Column<int> > dev_col = shared_pointer_namespace::static_pointer_cast<GPU_Typed_Base_Column<int> > (t1);
					shared_pointer_namespace::shared_ptr<GPU_Typed_Base_Column<int> > dev_col2 = shared_pointer_namespace::static_pointer_cast<GPU_Typed_Base_Column<int> > (t2);

					result_tids=my_sort_merge_join(dev_col,dev_col2);
				}else if(type==FLOAT){
					shared_pointer_namespace::shared_ptr<GPU_Typed_Base_Column<float> > dev_col = shared_pointer_namespace::static_pointer_cast<GPU_Typed_Base_Column<float> > (t1);
					shared_pointer_namespace::shared_ptr<GPU_Typed_Base_Column<float> > dev_col2 = shared_pointer_namespace::static_pointer_cast<GPU_Typed_Base_Column<float> > (t2);
					result_tids=my_sort_merge_join(dev_col,dev_col2);		
				}else if(type==VARCHAR){
					cout << "Fatal Error!" << "invalid AttributeType: " << type //<< " for Column: " << name 
			  			  << " Note: VARCHAR is currently not supported, will be added again in the future!"<< endl;
				}else if(type==BOOLEAN){
					//ptr=ColumnPtr(new Column<bool>(name,BOOLEAN));
					cout << "Fatal Error! invalid AttributeType: " << type //<<  " for Column: " << name 
						  << " Note: bool is currently not supported, will be added again in the future!"<< endl;
				}else{
					cout << "Fatal Error! invalid AttributeType: " << type << endl; //<< " for Column: " << name << endl;
			} 
			
				return result_tids;
			} 
			
	}; //end namespace gpu

}; //end namespace CoGaDB

